-module(tpTracerFile).

-export([
   start_link/2
   , init/3
]).

%% sys callbacks
-export([
   system_continue/3
   , system_terminate/4
   , system_code_change/4
]).

-record(state, {
   parent :: pid(),
   fileName :: file:filename_all(),
   size = 0 :: non_neg_integer(),
   maxSize :: infinity | non_neg_integer(),
   ioDevice :: file:io_device(),
   eventsPerFrame :: pos_integer(),
   eventsThisFrame = 0 :: non_neg_integer(),
   buffer = <<>> :: binary()
}).

start_link(Nth, Prefix) when is_list(Prefix) ->
   start_link(Nth, #{filename_prefix => Prefix});
start_link(Nth, Opts) when is_map(Opts) ->
   Pid = proc_lib:spawn_link(?MODULE, init, [self(), Nth, Opts]),
   {ok, Pid}.

init(Parent, Nth, Opts) ->
   %% Store all messages off the heap to avoid unnecessary GC.
   process_flag(message_queue_data, off_heap),
   %% We need to trap exit signals in order to shutdown properly.
   process_flag(trap_exit, true),
   %% No need to close the file, it'll be closed when the process exits.
   Filename = filename:flatten([maps:get(filename_prefix, Opts, "traces.lz4"), ".", integer_to_list(Nth)]),
   {ok, IoDevice} = file:open(Filename, [write, raw]),
   State = #state{parent = Parent, fileName = Filename, ioDevice = IoDevice, maxSize = maps:get(max_size, Opts, infinity), eventsPerFrame = maps:get(events_per_frame, Opts, 100000)},
   loop(State).

loop(State = #state{parent = Parent, size = Size, ioDevice = IoDevice, eventsPerFrame = MaxEvents, eventsThisFrame = NumEvents0, buffer = Buffer0}) ->
   receive
      {system, From, Request} ->
         sys:handle_system_msg(Request, From, Parent, ?MODULE, [], State);
      {'EXIT', Parent, Reason} ->
         terminate(Reason, State);
      Msg0 ->
         Msg = tpTerm:truncate(Msg0),
         Bin = term_to_binary(Msg),
         BinSize = byte_size(Bin),
         Buffer = <<Buffer0/binary, BinSize:32, Bin/binary>>,
         NumEvents = NumEvents0 + 1,
         if
            MaxEvents =:= NumEvents ->
               Frame = lz4f:compress_frame(Buffer),
               ok = file:write(IoDevice, Frame),
               maybe_rotate(State#state{size = Size + byte_size(Frame),
                  eventsThisFrame = 0, buffer = <<>>});
            true ->
               loop(State#state{eventsThisFrame = NumEvents, buffer = Buffer})
         end
   end.

%%IMY-todo 干掉这个函数 同时对于旋转的文件名 使用时间 年月日时分
maybe_rotate(State = #state{fileName = Filename, size = Size, maxSize = MaxSize,
   ioDevice = OldIoDevice}) when Size > MaxSize ->
   ok = file:close(OldIoDevice),
   ok = file:rename(Filename, Filename ++ ".bak"),
   {ok, NewIoDevice} = file:open(Filename, [write, raw]),
   loop(State#state{size = 0, ioDevice = NewIoDevice});
maybe_rotate(State) ->
   loop(State).

system_continue(_, _, State) ->
   loop(State).

-spec system_terminate(any(), _, _, #state{}) -> no_return().
system_terminate(Reason, _, _, State) ->
   terminate(Reason, State).

system_code_change(Misc, _, _, _) ->
   {ok, Misc}.

-spec terminate(any(), #state{}) -> no_return().
terminate(Reason, #state{ioDevice = IoDevice, buffer = Buffer}) ->
   _ = file:write(IoDevice, lz4f:compress_frame(Buffer)),
   exit(Reason).
